/*
 *	The RtTrend widget
 *
 *	Actually, two versions of the widget are defined here,
 *	and the DRAW_POLY macro can be defined to choose one of them.
 */

#ifdef _PHSLIB
	#include "/src/photon/lib/imports.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

#include <photon/realtime/RtTrend.h>

#if ! defined( DRAW_POLY )
	/*	DRAW_POLY can be 1
	 *	if PgDrawPolygon() should be used instead of PgDrawTrend().
	 *	Note that trend->lines[i].data will then point
	 *	to an array of PhPoint_t structures, not shorts.
	 */
	#define DRAW_POLY 0
#endif

#if ! DRAW_POLY && !defined Pg_TREND_PIXEL
	/* Remove this when PgDrawTrend() supports a 'pixel trend' flag */
	#define Pg_TREND_PIXEL 0
#endif

#ifndef _PHSLIB
	static PtWidgetClass_t *RtCreateTrendClass( void );
	static PtWidgetClassRef_t rt_trend = { NULL, RtCreateTrendClass };
	PtWidgetClassRef_t *RtTrend = &rt_trend;
#endif

#define Rt_TREND_FLAGS_PUBLIC		\
		(	Rt_PIXEL				\
		|	Rt_GRID					\
		|	Rt_TREND_HORIZONTAL		\
		|	Rt_TREND_VERTICAL		\
		|	Rt_TREND_LEFT_TO_RIGHT	\
		|	Rt_TREND_RIGHT_TO_LEFT	\
		|	Rt_TREND_TOP_TO_BOTTOM	\
		|	Rt_TREND_BOTTOM_TO_TOP	\
		|	Rt_GRID_ABOVE_TRENDS	\
		|	Rt_TRENDS_ABOVE_GRID	\
		|	Rt_GRID_IS_TRANSLUCENT	\
		|	Rt_GRID_FORCE			\
			)

#define NUM_PUBLIC_FLAGS 12

#if Rt_TREND_FLAGS_PUBLIC != ~( ~0u << NUM_PUBLIC_FLAGS )
	#error Some flags need to be adjusted
#endif

#define PRIVATE_FLAG( n )	( 1u << NUM_PUBLIC_FLAGS+(n) )

#define DONT_DRAW		PRIVATE_FLAG( 1 ) // Don't draw (there's no drivers)
#define DONT_BLIT		PRIVATE_FLAG( 2 ) // Redraw instead of blitting
#define DONT_MASKBLIT	PRIVATE_FLAG( 3 ) // Don't use pallette or masked blits
#define NO_GRID			PRIVATE_FLAG( 4 ) // Don't draw the grid
#define PALETTE_SENT	PRIVATE_FLAG( 5 ) // The palette has been sent to the driver
#define GRID_FLAGS	( NO_GRID | DONT_DRAW | DONT_BLIT | DONT_MASKBLIT )

#define FLAGS_INVALID	PRIVATE_FLAG( 0 )

#define Rt_TREND_FLAGS_PRIVATE	( FLAGS_INVALID | GRID_FLAGS | PALETTE_SENT )

/* Some flags for the grid */

#if	PALETTE_SENT == 0
	#error No room for the private flags
#endif

/*	Values for trend->modify */
#define START	3
#define DONE	2
#define DAMAGED	1
#define NEWDATA	0

static void reset_palette( RtTrendWidget_t *trend ) {
	/*
	 *	Set, clear, or just refresh the palette
	 */
	PhRid_t const c_rid = PgGetGC() -> rid;
	PgColor_t *list;
	switch ( trend->flags & ( PALETTE_SENT | DONT_MASKBLIT ) ) {
		default : // DONT_MASKBLIT
			/* We don't need a palette and don't have one */
			return;
		case PALETTE_SENT | DONT_MASKBLIT :
			/* We have a palette but don't need it */
			trend->flags &= ~PALETTE_SENT;
			PgSetRegion( 1 );
			PgSetPalette( NULL, 0, 0, -1,
				Pg_PALSET_HARDLOCKED | Pg_PALSET_FORCE_EXPOSE, 0
				);
			break;
		case PALETTE_SENT :
			/* We need a palette, but we've sent it already. Just refresh. */
			PgSetPalette( NULL, 0, 0, 0, Pg_PALSET_HARDLOCKED, 0 );
			return;
		case 0 :
			/* We need a palette, and we haven't sent it yet. */
			trend->flags |= PALETTE_SENT;
			PgSetRegion( 1 );
			PgSetPalette( list = trend->list, 0, trend->base, trend->color_count+1,
				Pg_PALSET_HARDLOCKED, 0
				);
			PgSetPalette( list + trend->grid_base, 0,
				trend->base + trend->grid_base, trend->color_count+1,
				Pg_PALSET_HARDLOCKED | Pg_PALSET_FORCE_EXPOSE, 0
				);
			}
	PgSetRegion( c_rid );
	}

static int RtTrend_Destroy( PtWidget_t *const widget ) {
	/*	Destroy callback: free all buffers */
	RtTrendWidget_t *const trend = (RtTrendWidget_t *)widget;
	RtTrendLines_t *line;
	if ( ( line = trend->lines ) != NULL ) {
		int i;
		for ( i = trend->trend_count; --i >= 0; ++line )
			free( line->data );
		}
	return Pt_CONTINUE;
	}

static int RtTrend_Unrealize( PtWidget_t *widget ) {
	RtTrendWidget_t *const trend = (RtTrendWidget_t *)widget;
	trend->flags |= DONT_MASKBLIT;
	reset_palette( trend );
	return Pt_CONTINUE;
	}

static void RtDfltsTrend( PtWidget_t *const widget ) {
	/*	Set defaults */
	RtTrendWidget_t *const trend = (RtTrendWidget_t *)widget;
	trend->basic.color      = Pg_RED;
	trend->basic.fill_color = Pg_BLACK;
	trend->grid_color		= Pg_GRAY;
	trend->flags
	=	Rt_TREND_HORIZONTAL | Rt_TREND_RIGHT_TO_LEFT
	|	Rt_GRID | Rt_GRID_IS_TRANSLUCENT;
	trend->min = SHRT_MIN;
	trend->max = SHRT_MAX;
	trend->width = USHRT_MAX;
	trend->grid_x =
	trend->grid_y = 5;
	trend->inc = 1;
	trend->pallen = 0x100;
	if	(	( trend->attrs = calloc( 1, sizeof(RtTrendAttr_t) ) ) != NULL
		&&	( trend->lines = calloc( 1, sizeof(RtTrendLines_t) ) ) != NULL
			)
		trend->attrs->map = trend->trend_count = 1;
	if ( ( trend->list = calloc( 4, sizeof(PgColor_t) ) ) != NULL ) {
		trend->color_count = 1;
		recalculate_palette( trend );
		}
	trend->modify = START;
	/*	Other fields are zeroed by default:
	 *	trend->trend_count = 0;			//	if calloc() failed
	 *	trend->lines[0].valid_size = 0;	//	if calloc() didn't fail
	 *	trend->list = NULL;
	 */
	}

static PgColor_t get_color( RtTrendWidget_t const *const trend, int const index ) {
	/*	Index is either in range 0...trend->colour_count
	 *	or equal to trend->grid_base.
	 *	Function converts the index to a colour value.
	 *	If the widget uses the palette, a palette index is returned.
	 *	If not, the colour value is taken from trend->list.
	 */
	#pragma aux get_color modify nomemory /* Enable some optimizations */
	PgColor_t color;
	if	(	( color = trend->list[ index ] ) != Pg_TRANSPARENT
		&&	!( trend->flags & DONT_MASKBLIT )
			)
		color = Pg_INDEX_COLOR | ( trend->base + index );
	return color;
	}

static void setgridflags( RtTrendWidget_t *trend ) {
	/*
	 *	Set private flags based on resources and system info
	 */
	unsigned flags = trend->flags & ~GRID_FLAGS;
	PhSysInfo_t si;
	if ( ! ( flags & Rt_GRID ) )
		flags |= DONT_MASKBLIT | NO_GRID;
	if ( PtQuerySystemInfo( (PtWidget_t*) trend, &si ) == NULL )
		flags |= DONT_BLIT | DONT_MASKBLIT;
	else if	(	si.gen.valid_fields & Ph_GEN_INFO_NUM_GFX
			&&	si.gen.num_gfx == 0
				)
		flags |= DONT_DRAW;
	else if	(	!( si.gfx.valid_fields & Ph_GFX_CAPABILITIES )
			||	!( si.gfx.capabilities & Ph_GCAP_BLIT )
				)
		flags |= DONT_MASKBLIT | DONT_BLIT;
	else if	(	(	si.gfx.capabilities
				&	( Ph_GCAP_MASKED_BLIT | Ph_GCAP_PALETTE )
					)
			!=	( Ph_GCAP_MASKED_BLIT | Ph_GCAP_PALETTE )
				)
		if ( flags & Rt_GRID_FORCE )
			flags |= DONT_BLIT | DONT_MASKBLIT;
		else
			flags |= NO_GRID | DONT_MASKBLIT;

	trend->flags = flags;
	}


/*
 *	RtDrawTrend is the draw function.
 *
 *	If there are some new data in the buffers, the function assumes
 *	that the trend is not damaged and only the new data must be drawn
 *	(and probably some old data must be shifted using the PgBlit() function).
 *	If there are no new data, the function redraws everything.
 */
static void RtDrawTrend( PtWidget_t *const widget, PhTile_t const *damage ) {
	RtTrendWidget_t	*trend = (RtTrendWidget_t*) widget;
	int tn, sample_size;
	PtWidget_t *wp;
	PhRect_t interior, drawrect;
	RtTrendLines_t *line;
	RtTrendAttr_t *attr;
	setgridflags( trend );

	if ( trend->flags & DONT_DRAW ) {
		trend->sample_size = 0;
		trend->modify = DONE;
		return;
		}

	(void) damage;

	if ( trend->flags & FLAGS_INVALID ) {
		/* We don't even know the direction! */
		PgSetFillColor( trend->basic.fill_color );
		PgDrawRect( PtWidgetCanvas( widget, &interior ), Pg_DRAW_FILL );
		PtBasicHighlight( widget );
		return;
		}

	reset_palette( trend );

	PtBasicInteriorExtent( widget, &interior );

	if ( trend->modify != NEWDATA )
		trend->flags |= DONT_BLIT;

	if ( trend->flags & DONT_BLIT ) {
		/* No new data or blitting not supported -- draw everything */
		sample_size = trend->n_data;
		/* We're  going to repair margins -- drawrect will include margins */
		PtWidgetCanvas( widget, &drawrect );
		}
	else {
		/*	We've got some new data.
		 *	As we assume that the widget isn't damaged,
		 *	we're not going to redraw the grid --
		 *	we'll just do a blit (unless there's nothing to blit)
		 *	and draw the new data.
		 *	This means that the plane mask can be set now.
		 */
		if ( !( trend->flags & DONT_MASKBLIT ) )
			PgSetPlaneMask( trend->tmask );
		if	(	( sample_size = trend->sample_size ) >= trend->n_data
			||	( wp = PtFindDiscontinuous( widget ) ) == NULL
				)
			/* No old data to blit or an error -- draw everything anyway */
			sample_size = trend->n_data;
		/* We're not going to repair margins -- interior can be copied to drawrect */
		drawrect = interior;
		}

	if ( sample_size == 0 ) {
		/*	Either trend->n_data == 0
		 *	or a bug ( trend->modify == NEWDATA but trend->sample_size == 0 )
		 */
		trend->modify = DONE;
		return;
		}

	/*	If we're going to draw all data and background,
	 *	drawrect is the rectangle to draw.
	 *	If not, we'll adjust drawrect to indicate the area
	 *	in which we will draw.
	 */
	if ( sample_size != trend->n_data ) {
		/*	We've got some old data to blit,
		 *	which means that only new data will be drawn.
		 *	We're going to adjust drawrect and do the blit.
		 *	sample_size is the blitting distance (measured in trend->inc units),
		 *	wp is the widget that owns the region.
		 */
		PhPoint_t blit, offs;
		PhRect_t blitrect = drawrect; /* The whole interior */
		if ( trend->flags & Rt_TREND_HORIZONTAL ) {
			if ( trend->flags & Rt_TREND_RIGHT_TO_LEFT ) {
				blitrect.ul.x -= blit.x = - sample_size * trend->inc;
				drawrect.ul.x = drawrect.lr.x + blit.x + 1;
				}
			else { /* LEFT_TO_RIGHT */
				blitrect.lr.x -= blit.x = sample_size * trend->inc;
				drawrect.lr.x = drawrect.ul.x + blit.x - 1;
				}
			blit.y = 0;
			}
		else {
			if ( trend->flags & Rt_TREND_BOTTOM_TO_TOP ) {
				blitrect.ul.y -= blit.y = - sample_size * trend->inc;
				drawrect.ul.y = drawrect.lr.y + blit.y + 1;
				}
			else {
				blitrect.lr.y -= blit.y = sample_size * trend->inc;
				drawrect.lr.y = drawrect.ul.y + blit.y - 1;
				}
			blit.x = 0;
			}

		PtTranslateRect( & blitrect, PtWidgetOffset( widget, &offs ) );
		PhBlit( wp->rid, &blitrect, &blit );
		PgFlush();
		}

	#if DRAW_POLY
		/* Erase drawrect */
		if ( !( trend->flags & DONT_MASKBLIT ) )
			PgSetPlaneMask( 0 );
		PgSetFillColor( get_color( trend, 0 ) );
		PgDrawRect( &drawrect, Pg_DRAW_FILL );

		/* Add clipping -- some samples can be outside of the range */
		PtClipAdd( widget, &interior );

		PgSetStrokeWidth( 0L );
		if ( !( trend->flags & NO_GRID ) ) {
			/*	Draw the grid. */
			PhPoint_t grid_lines;
			PgSetStrokeColor( get_color( trend, trend->grid_base ) );
			grid_lines.x = trend->grid_x + 1; /* because DrawGrid draws one less */
			grid_lines.y = trend->grid_y + 1;
			--interior.ul.x; --interior.ul.y;
			++interior.lr.x; ++interior.lr.y;
			PgDrawGrid( &interior, &grid_lines );
			if ( !( trend->flags & DONT_MASKBLIT ) )
				PgSetPlaneMask( trend->tmask );
			}

		/*	At last, we can draw the trends.
		 *	We try to draw sample_size+1 points
		 *	and let the last one be clipped off
		 */
		++sample_size;
		attr = trend->attrs;
		line = trend->lines;
		for ( tn = trend->trend_count; --tn >= 0; ++line, ++attr ) {
			int valid;
			if	(	( valid = sample_size ) <= line->valid_size
				||	( valid = line->valid_size ) != 0
					) {
				PgSetStrokeColor( get_color( trend, attr->map ) );
				if	(	valid == 1
					||	trend->flags & Rt_PIXEL
						)
					PgDrawPixelArray( (PhPoint_t*) line->data,
						valid, &interior.ul
						);
				else
					PgDrawPolygon( (PhPoint_t*) line->data,
						valid, &interior.ul, Pg_POLY_STROKE
						);
			}	}

	#else /* ! DRAW_POLY */

		/*	erase drawrect
		 *	by drawing a rectangle in the background color
		 *	OR redrawing the part of the old trend.
		 */
		PgSetStrokeWidth( 0L );
		if	(	trend->modify == NEWDATA
			&&	sample_size <= trend->n_data / 2
				) {
			/* Instead of drawing a rectangle, we will redraw the old data.
			 */

			/* Add clipping -- some samples can be outside of the range */
			PtClipAdd( widget, &interior );
			PgSetStrokeColor( get_color( trend, 0 ) );
			for ( line = trend->lines, tn = trend->trend_count; --tn >= 0; ++line ) {
				int valid, todraw;
				/*	We try to draw sample_size+1 points
				 *	and let them be clipped off
				 */
				if	(	( todraw = sample_size + 1 )
					<=	( valid = line->valid_size - sample_size )
					||	( todraw = valid ) > 1
						)
			  		PgDrawTrend(
						line->data + sample_size,
						&trend->origin, todraw, trend->delta, 0, 0, trend->dflag
						);
			}	}
		else {
			/*	We cannot 'undraw' the old data
			 *	because we either don't have enough old data in buffer
			 *	or don't assume that the widget isn't damaged.
			 *	Anyway, we must draw the rectangle.
			 */
			PgSetFillColor( get_color( trend, 0 ) );
			PgDrawRect( &drawrect, Pg_DRAW_FILL );
			PgFlush();
			/* Add clipping -- some samples can be outside of the range */
			PtClipAdd( widget, &interior );
			if ( !( trend->flags & NO_GRID ) ) {
				/*	Draw the grid. */
				PhPoint_t grid_lines;
				PgSetStrokeColor( get_color( trend, trend->grid_base ) );
				grid_lines.x = trend->grid_x + 1; /* because DrawGrid draws one less */
				grid_lines.y = trend->grid_y + 1;
				PgDrawGrid( &interior, &grid_lines );
				if ( !( trend->flags & DONT_MASKBLIT ) )
					PgSetPlaneMask( trend->tmask );
			}	}
		PgFlush();

		/*	At last, we can draw the trends.
		 *	We try to draw sample_size+1 points
		 *	and let the last one be clipped off
		 */
		++ sample_size;
		attr = trend->attrs;
		line = trend->lines;
		for ( tn = trend->trend_count; --tn >= 0; ++line, ++attr ) {
			int valid;
			if	(	( valid = sample_size ) <= line->valid_size
				||	( valid = line->valid_size ) > 1
					) {
				PgSetStrokeColor( get_color( trend, attr->map ) );
		  		PgDrawTrend(
					line->data,
					&trend->origin, valid, trend->delta, 0, 0, trend->dflag
					);
			}	}
		PgFlush();
	#endif /* DRAW_POLY */
	if ( ~( trend->flags & DONT_MASKBLIT ) )
		/*	Note that the plane mask has been set to trend->tmask
		 *	-- either before we did the blit
		 *	or after we drew the grid.
		 */

		PgSetPlaneMask( 0 );
	PtClipRemove();
	if ( trend->modify == NEWDATA )
		trend->sample_size = 0;
	else
		PtBasicHighlight( widget );

	trend->modify = DONE;
	}


/*
 *	average_colors gives the RGB value of the average of RGB values of the
 *	foreground and background color. This is used to generate the effect of
 *	a translucent grid lying over trends.
 */
static PgColor_t average_colors( PgColor_t const fg, PgColor_t const bg ) {
	#pragma aux average_colors modify nomemory parm nomemory
	const
		fg_r = PgRedValue( fg ),	bg_r = PgRedValue( bg ),
		fg_g = PgGreenValue( fg ),	bg_g = PgGreenValue( bg ),
		fg_b = PgBlueValue( fg ),	bg_b = PgBlueValue( bg );
	#if 1
		#define AVERAGE( a, b )	( ( a + b ) / 2 )
	#else
		/* use what's below if you want the grid to dominate */
		#define AVERAGE( a, b )	( a > b ? a : b )
	#endif
	return
		PgRGB(
			AVERAGE( fg_r, bg_r ),
			AVERAGE( fg_g, bg_g ),
			AVERAGE( fg_b, bg_b )
			);
	#undef AVERAGE
	}

static void rescale_data( RtTrendWidget_t *const trend,
		short newmin, short newmax, short newwidth
		) {
	/*
	 *	Rescale all valid data and set new min, max and width
	 */
	int tn;
	RtTrendLines_t *line = trend->lines;
	int const
		oldscale = trend->max - trend->min,
		newscale = newmax - newmin;
	for ( tn = trend->trend_count; --tn >= 0; ++line ) {
		int i;
		#if DRAW_POLY
			PhPoint_t *data = (PhPoint_t*) line->data;
			for ( i = line->valid_size; --i >= 0; ++data ) {
				int val
				=	trend->flags & Rt_TREND_HORIZONTAL
				?	data->y : data->x;
				/* Calculate old data value */
				val = trend->max - ( val * (long)oldscale ) / trend->width;
				/* Now calculate new point position */
				val = ( newwidth * (long)( newmax - val ) ) / newscale;
				if ( trend->flags & Rt_TREND_HORIZONTAL )
					data->y = val;
				else
					data->x = val;
				}
		#else
			short *data = line->data;
			for ( i = line->valid_size; --i >= 0; ++data ) {
				int const val /* The original data value (approximated) */
				= trend->max - ( *data * (long)oldscale ) / trend->width;
				*data = ( newwidth * (long)( newmax - val ) ) / newscale;
				}
		#endif
		}	
	trend->min = newmin;
	trend->max = newmax;
	trend->width = newwidth;
	}

static void recalculate_sizes( RtTrendWidget_t *const trend ) {
	/*	Call it after DIM or VERTICAL/HORIZONTAL flag changes */
	PhRect_t interior;
	long new_width;
	PtBasicWidgetCanvas( (PtWidget_t*)trend, &interior );
	#if DRAW_POLY
		new_width
		=	trend->flags & Rt_TREND_VERTICAL
		?	interior.lr.x - interior.ul.x + 1
 		:	interior.lr.y - interior.ul.y + 1;
	#else
		/*	Calculate trend->dflag, trend->delta, and trend->origin.
		 */
		trend->dflag = trend->flags & Rt_PIXEL ? Pg_TREND_PIXEL : 0;
		if ( trend->flags & Rt_TREND_HORIZONTAL ) {
			if ( trend->flags & Rt_TREND_RIGHT_TO_LEFT ) {
				trend->origin.x = interior.lr.x;
				trend->delta = - trend->inc;
				}
			else {
				trend->origin.x = interior.ul.x;
				trend->delta = trend->inc;
				}
			trend->origin.y = interior.ul.y;
			new_width = interior.lr.y - interior.ul.y + 1;
			trend->dflag |= Pg_TREND_HORIZ;
			}
		else {
			if ( trend->flags & Rt_TREND_BOTTOM_TO_TOP ) {
				trend->origin.y = interior.lr.y;
				trend->delta = - trend->inc;
				}
			else {
				trend->origin.y = interior.ul.y;
				trend->delta = trend->inc;
				}
			trend->origin.x = interior.ul.x;
			new_width = interior.lr.x - interior.ul.x + 1;
			trend->dflag |= Pg_TREND_VERT;
			}
	#endif /* DRAW_POLY */
	if ( new_width > 0 )
		rescale_data( trend, trend->min, trend->max, new_width );
		/* If new_width is 0 or negative, nothing will be drawn anyway! */
	}

static void discard_data( RtTrendWidget_t *const trend ) {
	/* Mark all data as invalid */
	RtTrendLines_t *line = trend->lines;
	int tn;
	for ( tn = trend->trend_count; --tn >= 0; ++line )
		line->valid_size = 0;
	}

static void set_range( RtTrendWidget_t *const trend, int const min, int const max ) {
	/*	Change the MIN or MAX resource */
	if ( max > min )
		rescale_data( trend, min, max, trend->width );
	else {
		discard_data( trend );
		trend->min = min;
		trend->max = max;
	}	}

static int RtSetTrend( RtTrendWidget_t *const trend, int start );

static void RtDamageTrend( RtTrendWidget_t *const trend ) {
	trend->modify = DAMAGED;
	trend->sample_size = 0; /* So that entire widget is redrawn */
	PtDamageWidget( (PtWidget_t*) trend );
	}

/*
 *	RtSetTrendCount is called whenever the number of trends is changed
 */
static void RtSetTrendCount( PtWidget_t *const widget, PtArg_t const *const arg ) {
	RtTrendWidget_t *const trend = (RtTrendWidget_t *)widget;
	RtTrendLines_t *line;
	RtTrendAttr_t *attr;
	int i, old_count;
	if ( ( i = arg->value ) < ( old_count = trend->trend_count ) )
		for	(	line = & trend->lines[ trend->trend_count = i ];
				i < old_count;
				++i, ++line
				)
			free( line->data );
	if	(	(	line
			=	realloc( trend->lines, sizeof(*line) * arg->value )
				)
		==	NULL
			)
		return;
	trend->lines = line;
	if	(	(	attr
			=	realloc( trend->attrs, sizeof(*attr) * arg->value )
				)
		==	NULL
			)
		return;
	trend->attrs = attr;
	trend->trend_count = arg->value;
	for	(	line += i = old_count, attr += i;
			i < trend->trend_count;
			++i, ++line, ++attr
			) {
		line->data = NULL; /* so that realloc doesn't screw up later */
		attr->map = 1; /* default */
		line->valid_size = 0;
		}
	RtSetTrend( trend, old_count );
	RtDamageTrend( trend );
	}

static unsigned grid_base( unsigned const colour_count ) {
	/*	The smallest power of two that is greater than colour_count
	 *	Assume that color_count is at least 1.
	 */
	#pragma aux grid_base modify nomemory parm nomemory
	unsigned i = 1;
	while ( ( i <<= 1 ) <= colour_count );
	return i;
	}

static void recalculate_palette( RtTrendWidget_t *const trend ) {
	if ( trend->color_count ) {
		int i, ii, j;
		PgColor_t *list = trend->list;
		list[ 0 ] = trend->basic.fill_color;
		list[ 1 ] = trend->basic.color;

		trend->grid_base = i = grid_base( trend->color_count );
		trend->tmask = i;
		list[ i ] = trend->grid_color;
		j = trend->pallen; ii = i+i;
		if ( ( j -= ii + j % ii ) < 0 )
			j = 0;
		trend->base = j;

		/*	This is what we have when trend->pallen is 0x100:
		 *
		 *	color_count		base	grid_base	tmask
		 *	1				252		2			0xFE
		 *	2...3			248		4			0xFC
		 *	4...7			240		8			0xF8
		 *	8...15			224		16			0xF0
		 *	16...31			192		32			0xE0
		 *	32...63			128		64			0xC0
		 */
		
		if ( trend->flags & Rt_GRID ) {
			/*	Set list[ grid_base+1 ... grid_base+color_count ]
			 *	acording to list[ 1...color_count ] and the grid flags
			 */
			i = trend->color_count;
			do	{
				++list;
				list[ trend->grid_base ]
				=	trend->flags & Rt_GRID_ABOVE_TRENDS
					?	trend->grid_color
					:	trend->flags & Rt_TRENDS_ABOVE_GRID
						?	*list
						:	average_colors( *list, trend->grid_color );
				}	while ( --i );
			}
		trend->flags &= ~PALETTE_SENT;
		RtDamageTrend( trend );
	}	}

/*
 *	RtTrendModifyPalette is called when either the list of colors used by
 *	RtTrend is modified, or the mapping of trends onto colors is modified.
 */
static void RtTrendModifyPalette( PtWidget_t *const widget, PtArg_t const *const arg ) {
	RtTrendWidget_t *const trend = (RtTrendWidget_t*) widget;
	PgColor_t *table;
	RtTrendAttr_t *attr;
	int i, j;

	switch( arg->type ) {
		case Rt_ARG_TREND_PALETTE_END :
			i = grid_base( trend->color_count ) << 1;
			j = trend->pallen;
			if ( j / i != ( trend->pallen = arg->value ) / i ) {
				trend->flags &= ~PALETTE_SENT;
				recalculate_palette( trend );
				}
			return;

		case Pt_ARG_FILL_COLOR : /* background color has changed */
			trend->basic.fill_color = arg->value;
			break;
		case Pt_ARG_COLOR : /* default foreground color */
			trend->basic.color = arg->value;
			break;
		case Rt_ARG_TREND_GRID_COLOR :
			trend->grid_color = arg->value;
			break;
		case Rt_ARG_TREND_COLOR_LIST : /* list of colors used */
			if ( ( arg->len < 1 ) || ( arg->len > 63 ) ) {
				fputs( "RtTrend: color count must be between 1 and 63\n", stderr );
				return;
				}

			if	(	(	table
					=	realloc( trend->list,
							grid_base( arg->len ) * 2 * sizeof(PgColor_t)
						)	)
				==	NULL
					)
				return;

			trend->list = table;
			memcpy(
				++table,
				(void const*) arg->value,
				( trend->color_count = arg->len ) * sizeof(*table)
				);
			trend->basic.color = *table;

			break;
		case Rt_ARG_TREND_ATTRIBUTES : /* mapping of trends onto colors */
			/*	It seems that the mapping of trend 0 cannot be set independently */
			if ( arg->len >= trend->trend_count )
				/*	Either trend_count == 0
				 *	or arg->len is an invalid trend number
				 */
				return;
			i = sizeof(*attr);
			attr = trend->attrs;
			if ( arg->len == 0 ) /* map all trends */
				i *= trend->trend_count;
			else /* map trend arg->len */
				attr += arg->len;
			memcpy( attr, (void *) arg->value, i );
			break;
		}
	for ( attr = trend->attrs, i = trend->trend_count; --i >= 0; ++attr )
		if ( attr->map > trend->color_count )
			attr->map = 1;
	if ( arg->type != Rt_ARG_TREND_ATTRIBUTES )
		recalculate_palette( trend );
	RtDamageTrend( trend );
	}

static unsigned long validate_flags( unsigned long const flags ) {
	/*	Add FLAGS_INVALID if the flags are invalid */
	#pragma aux validate_flags modify nomemory parm nomemory
	struct FlagSets {
		unsigned long	IfSet,	/* If all these flags are set, ... */
						Set;	/* ... then exactly one of these should be set */
		};
	static const struct FlagSets
		flagsets[] = {
			{	0,
				Rt_TREND_HORIZONTAL | Rt_TREND_VERTICAL
				},
			{	Rt_TREND_HORIZONTAL,
				Rt_TREND_RIGHT_TO_LEFT | Rt_TREND_LEFT_TO_RIGHT
				},
			{	Rt_TREND_VERTICAL,
				Rt_TREND_BOTTOM_TO_TOP | Rt_TREND_TOP_TO_BOTTOM
				},
			{	Rt_GRID,
				Rt_GRID_ABOVE_TRENDS | Rt_TRENDS_ABOVE_GRID | Rt_GRID_IS_TRANSLUCENT
				}
			};
	const struct FlagSets *ptr = flagsets;
	do	if ( ( flags & ptr->IfSet ) == ptr->IfSet ) {
			unsigned long bits = flags & ptr->Set;
			int n = 0;
			while ( bits ) {
				if ( bits & 1 )
					++n;
				bits >>= 1;
				}
			if ( n != 1 )
				return flags | FLAGS_INVALID;
			}
		while ( ++ptr != & flagsets[ sizeof(flagsets) / sizeof(flagsets[0]) ] );
	return flags;
	}

static void RtModifyTrend( PtWidget_t *const widget, PtArg_t const *const arg ) {
	RtTrendWidget_t *trend = (RtTrendWidget_t*) widget;
	long flags, mask;

	switch ( arg->type ) {
		case Rt_ARG_TREND_GRID_X :
			trend->grid_x = arg->value;
			break;
		case Rt_ARG_TREND_GRID_Y :
			trend->grid_y = arg->value;
			break;
		case Rt_ARG_TREND_FLAGS :
			mask = arg->len & ~Rt_TREND_FLAGS_PRIVATE;
			if	(	(	(	trend->flags
						=	validate_flags(
								arg->value & mask
								|	( flags = trend->flags ) & ~mask
							)	)
					&	FLAGS_INVALID
						)
				==	0
					) {
				/*	New flags are valid */
				RtSetTrend( trend, 0 );
				if	(	( mask = flags ^ trend->flags )
					&	(	Rt_TREND_HORIZONTAL		|	Rt_TREND_VERTICAL
						|	Rt_TREND_LEFT_TO_RIGHT	|	Rt_TREND_RIGHT_TO_LEFT
						|	Rt_TREND_TOP_TO_BOTTOM	|	Rt_TREND_BOTTOM_TO_TOP
						|	FLAGS_INVALID
						)	)
					recalculate_sizes( trend );
				setgridflags( trend );
				if	(	trend->color_count > 0
					&&	mask
					&	(	Rt_GRID
						|	Rt_GRID_ABOVE_TRENDS
						|	Rt_TRENDS_ABOVE_GRID
						|	Rt_GRID_IS_TRANSLUCENT
						|	FLAGS_INVALID
						)	)
					/*	If these flags have been changed,
					 *	the palette must be recalculated.
					 */
					recalculate_palette( trend );
				}
			break;
		case Rt_ARG_TREND_MAX:
			set_range( trend, trend->min, arg->value );
			break;
		case Rt_ARG_TREND_MIN:
			set_range( trend, arg->value, trend->max );
			break;
		case Rt_ARG_TREND_INC:
			if ( arg->value <= 0 )
				return;
			trend->inc = arg->value;
			RtSetTrend( trend, 0 );
			recalculate_sizes( trend );
			break;
		case Rt_ARG_TREND_DATA:
			if ( arg->value ) {
				RtTrendPutData( widget, (short*) arg->value, arg->len );
				return;
				}
			else
				discard_data( trend );
			break;
		}
	RtDamageTrend( trend );
	}

static int RtSetTrend( RtTrendWidget_t *trend, int start ) {
	PhRect_t		rect;
	RtTrendLines_t *line;

	int length;

	PtBasicWidgetCanvas( (PtWidget_t*)trend, &rect);

	length
	=	trend->flags & Rt_TREND_HORIZONTAL
	?	rect.lr.x - rect.ul.x
	:	rect.lr.y - rect.ul.y;

	if ( length < 0 )
		length = 0;

	trend->n_data = length / trend->inc + 1;

	if ( trend->sample_size > trend->n_data )
		trend->sample_size = trend->n_data;

	for	(	line = &trend->lines[ start ];
			start < trend->trend_count;
			++start, ++line
			) {
		#if DRAW_POLY
			PhPoint_t *data;
		#else
			short *data;
		#endif
		if ( line->valid_size > trend->n_data )
			line->valid_size = trend->n_data;
		if	(	(	data
				=	realloc( line->data, sizeof(*data) * trend->n_data )
					)
			!=	NULL
				) {
			#if DRAW_POLY
				line->data = (void*) data;
				if ( trend->flags & Rt_TREND_VERTICAL ) {
					int i, y, delta;
					if ( trend->flags & Rt_TREND_BOTTOM_TO_TOP ) {
						y = trend->inc * ( trend->n_data - 1 );
						delta = - trend->inc;
						}
					else {
						y = 0;
						delta = trend->inc;
						}
					for ( i = trend->n_data; --i >= 0; y += delta )
						data++->y = y;
					}
				else {
					int i, x, delta;
					if ( trend->flags & Rt_TREND_RIGHT_TO_LEFT ) {
						x = trend->inc * ( trend->n_data - 1 );
						delta = - trend->inc;
						}
					else {
						x = 0;
						delta = trend->inc;
						}
					for ( i = trend->n_data; --i >= 0; x += delta )
						data++->x = x;
					}
			#else /* ! DRAW_POLY */
				line->data = data;
			#endif /* DRAW_POLY */
			}
		else {
			/*	Emergency -- reduce the trend count */
			do	if ( ( data = (void*) line++->data ) != NULL )
					free( data );
				while ( --trend->trend_count > start );
			return -1;
		}	}
	return Pt_CONTINUE;
	}

void RtTrendPutData( PtWidget_t *widget, short const *newdata, int data_len ) {
	RtTrendWidget_t *trend = (RtTrendWidget_t*) widget;
	int i, tn, old, sample_size;
	long scale;
	RtTrendLines_t *line;

	if	(	trend->flags & FLAGS_INVALID
		||	( sample_size = data_len / trend->trend_count ) <= 0
			)
		return;
	/*	If the widget has been drawn since RtTrendPutData was last called,
	 *	trend->sample_size is now zero.
	 *	Otherwise trend->sample_size is the number of samples that have not been drawn.
	 */
	if ( widget->flags & Pt_DAMAGED ) {
		/* If the widget is damaged, we need to redraw it anyway */
		trend->sample_size = 0;
		trend->modify = DAMAGED;
		}
	else {
		trend->sample_size += sample_size;
		trend->modify = NEWDATA;
		}
	if ( ( scale = trend->max - trend->min ) <= 0 )
		return;
	line = trend->lines;
	#if DRAW_POLY
		old = trend->n_data - sample_size;
		if ( trend->flags & Rt_TREND_VERTICAL ) {
			for ( tn = trend->trend_count; --tn >= 0; ++line ) {
				PhPoint_t *dst = (PhPoint_t*)line->data + trend->n_data;
				if ( ( i = old ) > 0 ) {
					PhPoint_t *src = dst - sample_size;
					do	(--dst)->x = (--src)->x;
						while ( --i );
					i = sample_size;
					}
				else {
					newdata -= i; /* i is negative */
					i = trend->n_data;
					}
				do	(--dst)->x = trend->width * ( (long)trend->max - *newdata++ ) / scale;
					while ( --i );
				if ( ( line->valid_size += sample_size ) > trend->n_data )
					line->valid_size = trend->n_data;
			}	}
		else {
			for ( tn = trend->trend_count; --tn >= 0; ++line ) {
				PhPoint_t *dst = (PhPoint_t*)line->data + trend->n_data;
				if ( ( i = old ) > 0 ) {
					PhPoint_t *src = dst - sample_size;
					do	(--dst)->y = (--src)->y;
						while ( --i );
					i = sample_size;
					}
				else {
					newdata -= i; /* i is negative */
					i = trend->n_data;
					}
				do	(--dst)->y = trend->width * ( (long)trend->max - *newdata++ ) / scale;
					while ( --i );
				if ( ( line->valid_size += sample_size ) > trend->n_data )
					line->valid_size = trend->n_data;
			}	}
	#else /* ! DRAW_POLY */
		if ( ( old = trend->n_data - sample_size ) > 0 ) {
			old *= sizeof(line->data[0]);
			for ( tn = trend->trend_count; --tn >= 0; ++line ) {
				short *dst = line->data;
				dst = memmove( dst + ( i = sample_size ), dst, old );
				do	*--dst = trend->width * ( (long)trend->max - *newdata++ ) / scale;
					while ( --i );
				if ( ( line->valid_size += sample_size ) > trend->n_data )
					line->valid_size = trend->n_data;
			}	}
		else
			for ( tn = trend->trend_count; --tn >= 0; ++line ) {
				short *dst = line->data + ( i = trend->n_data );
				newdata -= old;  /* 'old' is negative, remember? */
				do	*--dst = trend->width * ( (long)trend->max - *newdata++ ) / scale;
					while ( --i );
				if ( ( line->valid_size += sample_size ) > trend->n_data )
					line->valid_size = trend->n_data;
				}
	#endif /* DRAW_POLY */
	PtDamageWidget( widget );
	}

static void RtExtentTrend( PtWidget_t *widget ) {
	RtTrendWidget_t *const trend = (RtTrendWidget_t*) widget;
	PtSuperClassExtent( PtBasic, widget );
	RtSetTrend( trend, 0 );
	recalculate_sizes( trend );
	}

PtWidgetClass_t *RtCreateTrendClass( void ) {
	static const PtResourceRec_t resources[] = {
		{	Pt_ARG_COLOR,			RtTrendModifyPalette, NULL,
			Pt_ARG_MODE_LONG | offsetof( PtBasicWidget_t, color )
			},
		{	Pt_ARG_FILL_COLOR,		RtTrendModifyPalette, NULL,
			Pt_ARG_MODE_LONG | offsetof( PtBasicWidget_t, fill_color )
			},
		{	Rt_ARG_TREND_DATA,		RtModifyTrend, NULL,
			Pt_ARG_MODE_ALLOC 	| offsetof( RtTrendWidget_t, lines)
			},
		{	Rt_ARG_TREND_GRID_X,		RtModifyTrend, NULL,
			Pt_ARG_MODE_SHORT 	| offsetof(RtTrendWidget_t, grid_x)
			},
		{	Rt_ARG_TREND_GRID_Y,		RtModifyTrend, NULL,
			Pt_ARG_MODE_SHORT 	| offsetof(RtTrendWidget_t, grid_y)
			},
		{	Rt_ARG_TREND_COUNT,		RtSetTrendCount, NULL,
			Pt_ARG_MODE_SHORT 	| offsetof(RtTrendWidget_t, trend_count)
			},
		{	Rt_ARG_TREND_MAX,		RtModifyTrend, NULL,
			Pt_ARG_MODE_SHORT 	| offsetof(RtTrendWidget_t, max)
			},
		{	Rt_ARG_TREND_MIN,		RtModifyTrend, NULL,
			Pt_ARG_MODE_SHORT 	| offsetof(RtTrendWidget_t, min)
			},
		{	Rt_ARG_TREND_INC,		RtModifyTrend, NULL,
			Pt_ARG_MODE_SHORT 	| offsetof(RtTrendWidget_t, inc)
			},
		{	Rt_ARG_TREND_FLAGS,		RtModifyTrend, NULL,
			Pt_ARG_MODE_LONG_FLAGS  | offsetof(RtTrendWidget_t,flags)
			},
  		{	Rt_ARG_TREND_GRID_COLOR,	RtTrendModifyPalette, NULL,
  			Pt_ARG_MODE_LONG	| offsetof(RtTrendWidget_t,grid_color)
			},
		{	Rt_ARG_TREND_COLOR_LIST,	RtTrendModifyPalette, NULL,
			Pt_ARG_MODE_ARRAY	| offsetof( RtTrendWidget_t, list ),
			Pt_ARG_MODE_SHORT	| offsetof( RtTrendWidget_t, color_count )
			},
		{	Rt_ARG_TREND_ATTRIBUTES, 	RtTrendModifyPalette, NULL,
			Pt_ARG_MODE_ARRAY	| offsetof( RtTrendWidget_t, attrs ),
			Pt_ARG_MODE_SHORT	| offsetof( RtTrendWidget_t, trend_count )
			},
		{	Rt_ARG_TREND_PALETTE_END, RtTrendModifyPalette, NULL,
			Pt_ARG_MODE_SHORT	| offsetof( RtTrendWidget_t, pallen )
		}	};

	static const PtArg_t args[] = {
		{ Pt_SET_VERSION, 110 },
		{ Pt_SET_RESOURCES,		(long) resources, sizeof( resources )/sizeof( resources[0] ) },
		{ Pt_SET_NUM_RESOURCES, sizeof( resources )/sizeof( resources[0] ) },
		{ Pt_SET_STATE_LEN,		sizeof( RtTrendWidget_t ) },
		{ Pt_SET_DFLTS_F,		(long) RtDfltsTrend },
		{ Pt_SET_EXTENT_F,		(long) RtExtentTrend },
		{ Pt_SET_DRAW_F,		(long) RtDrawTrend },
		{ Pt_SET_DESTROY_F,		(long) RtTrend_Destroy },
		{ Pt_SET_UNREALIZE_F,	(long) RtTrend_Unrealize },
		{ Pt_SET_FLAGS,			Pt_RECTANGULAR, Pt_RECTANGULAR },
		};
		
	return( RtTrend->wclass = PtCreateWidgetClass(
		PtBasic, 0,	sizeof( args )/sizeof( args[0] ), args ) );
	}

